#include "Estimator.h"

Estimator::Estimator(const float& filter_corner, const float& filter_surf,
                     int ivox_nearby_type,int max_ivox_node,float resolution){

  laserCloudCornerLast.resize(SLIDEWINDOWSIZE);
  for(auto& p:laserCloudCornerLast)
    p.reset(new pcl::PointCloud<PointType>);
  laserCloudSurfLast.resize(SLIDEWINDOWSIZE);
  for(auto& p:laserCloudSurfLast)
    p.reset(new pcl::PointCloud<PointType>);
  laserCloudNonFeatureLast.resize(SLIDEWINDOWSIZE);
  for(auto& p:laserCloudNonFeatureLast)
    p.reset(new pcl::PointCloud<PointType>);
  laserCloudCornerStack.resize(SLIDEWINDOWSIZE);
  for(auto& p:laserCloudCornerStack)
    p.reset(new pcl::PointCloud<PointType>);
  laserCloudSurfStack.resize(SLIDEWINDOWSIZE);
  for(auto& p:laserCloudSurfStack)
    p.reset(new pcl::PointCloud<PointType>);
  laserCloudNonFeatureStack.resize(SLIDEWINDOWSIZE);
  for(auto& p:laserCloudNonFeatureStack)
    p.reset(new pcl::PointCloud<PointType>);
  laserCloudCornerForMap.reset(new pcl::PointCloud<PointType>);
  laserCloudSurfForMap.reset(new pcl::PointCloud<PointType>);
  laserCloudNonFeatureForMap.reset(new pcl::PointCloud<PointType>);
  transformForMap.setIdentity();


  /*for(int i = 0; i < localMapWindowSize; i++){
    localCornerMap[i].reset(new pcl::PointCloud<PointType>);
    localSurfMap[i].reset(new pcl::PointCloud<PointType>);
    localNonFeatureMap[i].reset(new pcl::PointCloud<PointType>);
  }*/

  downSizeFilterCorner.setLeafSize(filter_corner, filter_corner, filter_corner);
  downSizeFilterSurf.setLeafSize(filter_surf, filter_surf, filter_surf);
  downSizeFilterNonFeature.setLeafSize(0.4, 0.4, 0.4);

    //local_map_.Init(ivox_nearby_type,max_ivox_node,resolution);
}

void Estimator::print()
{

}
Estimator::~Estimator(){

}

void Estimator::processPointToLine(std::vector<ceres::CostFunction *>& edges,
                                   std::vector<FeatureLine>& vLineFeatures,
                                   const pcl::PointCloud<PointType>::Ptr& laserCloudCorner,
                                   const Eigen::Matrix4d& exTlb,
                                   const Eigen::Matrix4d& m4d){

  Eigen::Matrix4d Tbl = Eigen::Matrix4d::Identity();
  Tbl.topLeftCorner(3,3) = exTlb.topLeftCorner(3,3).transpose();
  Tbl.topRightCorner(3,1) = -1.0 * Tbl.topLeftCorner(3,3) * exTlb.topRightCorner(3,1);
  if(!vLineFeatures.empty())
  {
    for(const auto& l : vLineFeatures){
      auto* e = Cost_NavState_IMU_Line::Create(l.pointOri,
                                               l.lineP1,
                                               l.lineP2,
                                               Tbl,
                                               Eigen::Matrix<double, 1, 1>(1/IMUIntegrator::lidar_m));
      edges.push_back(e);
    }
    return;
  }
  PointType _pointOri, _pointSel, _coeff;
  std::vector<int> _pointSearchInd;
  std::vector<float> _pointSearchSqDis;
  std::vector<int> _pointSearchInd2;
  std::vector<float> _pointSearchSqDis2;

  Eigen::Matrix< double, 3, 3 > _matA1;
  _matA1.setZero();

  int laserCloudCornerStackNum = laserCloudCorner->points.size();
  pcl::PointCloud<PointType>::Ptr kd_pointcloud(new pcl::PointCloud<PointType>);
  int debug_num1 = 0;
  int debug_num2 = 0;
  int debug_num12 = 0;
  int debug_num22 = 0;
  for (int i = 0; i < laserCloudCornerStackNum; i++)
  {
    _pointOri = laserCloudCorner->points[i];
      LocalMapIvox::pointAssociateToMap(&_pointOri, &_pointSel, m4d);

    if(std::isnan(_pointSel.x) || std::isnan(_pointSel.y) ||std::isnan(_pointSel.z)) continue;

    if(m_global_map.laserCloudCornerFromLocal->points.size() > 100)
    {
      m_global_map.kdtreeCornerFromLocal->nearestKSearch(_pointSel, 5, _pointSearchInd, _pointSearchSqDis);

      if (_pointSearchSqDis[4] < thres_dist)
      {

        debug_num1++;
        float cx = 0;
        float cy = 0;
        float cz = 0;
        for (int j = 0; j < 5; j++)
        {
          cx += m_global_map.laserCloudCornerFromLocal->points[_pointSearchInd[j]].x;
          cy += m_global_map.laserCloudCornerFromLocal->points[_pointSearchInd[j]].y;
          cz += m_global_map.laserCloudCornerFromLocal->points[_pointSearchInd[j]].z;
        }
        cx /= 5;
        cy /= 5;
        cz /= 5;

        float a11 = 0;
        float a12 = 0;
        float a13 = 0;
        float a22 = 0;
        float a23 = 0;
        float a33 = 0;
        for (int j = 0; j < 5; j++)
        {
          float ax = m_global_map.laserCloudCornerFromLocal->points[_pointSearchInd[j]].x - cx;
          float ay = m_global_map.laserCloudCornerFromLocal->points[_pointSearchInd[j]].y - cy;
          float az = m_global_map.laserCloudCornerFromLocal->points[_pointSearchInd[j]].z - cz;

          a11 += ax * ax;
          a12 += ax * ay;
          a13 += ax * az;
          a22 += ay * ay;
          a23 += ay * az;
          a33 += az * az;
        }
        a11 /= 5;
        a12 /= 5;
        a13 /= 5;
        a22 /= 5;
        a23 /= 5;
        a33 /= 5;

        _matA1(0, 0) = a11;
        _matA1(0, 1) = a12;
        _matA1(0, 2) = a13;
        _matA1(1, 0) = a12;
        _matA1(1, 1) = a22;
        _matA1(1, 2) = a23;
        _matA1(2, 0) = a13;
        _matA1(2, 1) = a23;
        _matA1(2, 2) = a33;

        Eigen::SelfAdjointEigenSolver<Eigen::Matrix3d> saes(_matA1);
        Eigen::Vector3d unit_direction = saes.eigenvectors().col(2);

        if (saes.eigenvalues()[2] > 3 * saes.eigenvalues()[1])
        {
          debug_num12++;
          float x1 = cx + 0.1 * unit_direction[0];
          float y1 = cy + 0.1 * unit_direction[1];
          float z1 = cz + 0.1 * unit_direction[2];
          float x2 = cx - 0.1 * unit_direction[0];
          float y2 = cy - 0.1 * unit_direction[1];
          float z2 = cz - 0.1 * unit_direction[2];

          Eigen::Vector3d tripod1(x1, y1, z1);
          Eigen::Vector3d tripod2(x2, y2, z2);
          auto *e = Cost_NavState_IMU_Line::Create(Eigen::Vector3d(_pointOri.x, _pointOri.y, _pointOri.z),
                                                   tripod1,
                                                   tripod2,
                                                   Tbl,
                                                   Eigen::Matrix<double, 1, 1>(1 / IMUIntegrator::lidar_m));
          edges.push_back(e);
          vLineFeatures.emplace_back(Eigen::Vector3d(_pointOri.x, _pointOri.y, _pointOri.z),
                                     tripod1,
                                     tripod2);
          vLineFeatures.back().ComputeError(m4d);

          continue;
        }

      }

    }

    /*if(local_map_.CornerMapVox()->NumPoints()>20 )
    {
        pcl::PointCloud<PointType> nearst;
        local_map_.CornerMapVox()->GetClosestPoint(_pointSel, nearst.points, 5, thres_dist);

        if (nearst.size() == 5)
        {

            debug_num2++;
            float cx = 0;
            float cy = 0;
            float cz = 0;
            for (int j = 0; j < 5; j++)
            {
                cx += nearst.points[j].x;
                cy += nearst.points[j].y;
                cz += nearst.points[j].z;
            }
            cx /= 5;
            cy /= 5;
            cz /= 5;

            float a11 = 0;
            float a12 = 0;
            float a13 = 0;
            float a22 = 0;
            float a23 = 0;
            float a33 = 0;
            for (int j = 0; j < 5; j++)
            {
                float ax = nearst.points[j].x - cx;
                float ay = nearst.points[j].y - cy;
                float az = nearst.points[j].z - cz;

                a11 += ax * ax;
                a12 += ax * ay;
                a13 += ax * az;
                a22 += ay * ay;
                a23 += ay * az;
                a33 += az * az;
            }
            a11 /= 5;
            a12 /= 5;
            a13 /= 5;
            a22 /= 5;
            a23 /= 5;
            a33 /= 5;

            _matA1(0, 0) = a11;
            _matA1(0, 1) = a12;
            _matA1(0, 2) = a13;
            _matA1(1, 0) = a12;
            _matA1(1, 1) = a22;
            _matA1(1, 2) = a23;
            _matA1(2, 0) = a13;
            _matA1(2, 1) = a23;
            _matA1(2, 2) = a33;

            Eigen::SelfAdjointEigenSolver<Eigen::Matrix3d> saes(_matA1);
            Eigen::Vector3d unit_direction = saes.eigenvectors().col(2);

            if (saes.eigenvalues()[2] > 3 * saes.eigenvalues()[1])
            {
                debug_num22++;
                float x1 = cx + 0.1 * unit_direction[0];
                float y1 = cy + 0.1 * unit_direction[1];
                float z1 = cz + 0.1 * unit_direction[2];
                float x2 = cx - 0.1 * unit_direction[0];
                float y2 = cy - 0.1 * unit_direction[1];
                float z2 = cz - 0.1 * unit_direction[2];

                Eigen::Vector3d tripod1(x1, y1, z1);
                Eigen::Vector3d tripod2(x2, y2, z2);
                auto *e = Cost_NavState_IMU_Line::Create(Eigen::Vector3d(_pointOri.x, _pointOri.y, _pointOri.z),
                                                         tripod1,
                                                         tripod2,
                                                         Tbl,
                                                         Eigen::Matrix<double, 1, 1>(1 / IMUIntegrator::lidar_m));
                edges.push_back(e);
                vLineFeatures.emplace_back(Eigen::Vector3d(_pointOri.x, _pointOri.y, _pointOri.z),
                                           tripod1,
                                           tripod2);
                vLineFeatures.back().ComputeError(m4d);
            }
        }
    }*/
     
  }
}


void Estimator::processPointToPlanVec(std::vector<ceres::CostFunction *>& edges,
                                   std::vector<FeaturePlanVec>& vPlanFeatures,
                                   const pcl::PointCloud<PointType>::Ptr& laserCloudSurf,
                                   const Eigen::Matrix4d& exTlb,
                                   const Eigen::Matrix4d& m4d){
  Eigen::Matrix4d Tbl = Eigen::Matrix4d::Identity();
  Tbl.topLeftCorner(3,3) = exTlb.topLeftCorner(3,3).transpose();
  Tbl.topRightCorner(3,1) = -1.0 * Tbl.topLeftCorner(3,3) * exTlb.topRightCorner(3,1);
  if(!vPlanFeatures.empty()){
    for(const auto& p : vPlanFeatures)
    {
      auto* e = Cost_NavState_IMU_Plan_Vec::Create(p.pointOri,
                                                   p.pointProj,
                                                   Tbl,
                                                   p.sqrt_info);
      edges.push_back(e);
    }
    return;
  }
  PointType _pointOri, _pointSel, _coeff;
  std::vector<int> _pointSearchInd;
  std::vector<float> _pointSearchSqDis;
  std::vector<int> _pointSearchInd2;
  std::vector<float> _pointSearchSqDis2;

  Eigen::Matrix< double, 5, 3 > _matA0;
  _matA0.setZero();
  Eigen::Matrix< double, 5, 1 > _matB0;
  _matB0.setOnes();
  _matB0 *= -1;
  Eigen::Matrix< double, 3, 1 > _matX0;
  _matX0.setZero();
  int laserCloudSurfStackNum = laserCloudSurf->points.size();

  int debug_num1 = 0;
  int debug_num2 = 0;
  int debug_num12 = 0;
  int debug_num22 = 0;
  for (int i = 0; i < laserCloudSurfStackNum; i++) {
    _pointOri = laserCloudSurf->points[i];
      LocalMapIvox::pointAssociateToMap(&_pointOri, &_pointSel, m4d);

    if(std::isnan(_pointSel.x) || std::isnan(_pointSel.y) ||std::isnan(_pointSel.z)) continue;

    if(m_global_map.laserCloudSurfFromLocal->points.size() > 50) {
      m_global_map.kdtreeSurfFromLocal->nearestKSearch(_pointSel, 5, _pointSearchInd, _pointSearchSqDis);

      if (_pointSearchSqDis[4] < thres_dist) {
        debug_num1 ++;
        for (int j = 0; j < 5; j++) {
          _matA0(j, 0) = m_global_map.laserCloudSurfFromLocal->points[_pointSearchInd[j]].x;
          _matA0(j, 1) = m_global_map.laserCloudSurfFromLocal->points[_pointSearchInd[j]].y;
          _matA0(j, 2) = m_global_map.laserCloudSurfFromLocal->points[_pointSearchInd[j]].z;
        }
        _matX0 = _matA0.colPivHouseholderQr().solve(_matB0);

        float pa = _matX0(0, 0);
        float pb = _matX0(1, 0);
        float pc = _matX0(2, 0);
        float pd = 1;

        float ps = std::sqrt(pa * pa + pb * pb + pc * pc);
        pa /= ps;
        pb /= ps;
        pc /= ps;
        pd /= ps;

        bool planeValid = true;
        for (int j = 0; j < 5; j++) {
          if (std::fabs(pa * m_global_map.laserCloudSurfFromLocal->points[_pointSearchInd[j]].x +
                        pb * m_global_map.laserCloudSurfFromLocal->points[_pointSearchInd[j]].y +
                        pc * m_global_map.laserCloudSurfFromLocal->points[_pointSearchInd[j]].z + pd) > 0.2) {
            planeValid = false;
            break;
          }
        }

        if (planeValid) {
          debug_num12 ++;
          double dist = pa * _pointSel.x +
                        pb * _pointSel.y +
                        pc * _pointSel.z + pd;
          Eigen::Vector3d omega(pa, pb, pc);
          Eigen::Vector3d point_proj = Eigen::Vector3d(_pointSel.x,_pointSel.y,_pointSel.z) - (dist * omega);
          Eigen::Vector3d e1(1, 0, 0);
          Eigen::Matrix3d J = e1 * omega.transpose();
          Eigen::JacobiSVD<Eigen::Matrix3d> svd(J, Eigen::ComputeThinU | Eigen::ComputeThinV);
          Eigen::Matrix3d R_svd = svd.matrixV() * svd.matrixU().transpose();
          Eigen::Matrix3d info = (1.0/IMUIntegrator::lidar_m) * Eigen::Matrix3d::Identity();
          info(1, 1) *= plan_weight_tan;
          info(2, 2) *= plan_weight_tan;
          Eigen::Matrix3d sqrt_info = info * R_svd.transpose();

          auto* e = Cost_NavState_IMU_Plan_Vec::Create(Eigen::Vector3d(_pointOri.x,_pointOri.y,_pointOri.z),
                                                       point_proj,
                                                       Tbl,
                                                       sqrt_info);
          edges.push_back(e);
          vPlanFeatures.emplace_back(Eigen::Vector3d(_pointOri.x,_pointOri.y,_pointOri.z),
                                     point_proj,
                                     sqrt_info);
          vPlanFeatures.back().ComputeError(m4d);

          continue;
        }
        
      }
    }


    /*if(local_map_.surfMapVox()->NumPoints() > 20 )
    {
        pcl::PointCloud<PointType> nearst;
        local_map_.surfMapVox()->GetClosestPoint(_pointSel, nearst.points, 5, thres_dist);
        if (nearst.size() == 5)
        {
            debug_num2++;
            for (int j = 0; j < 5; j++)
            {
                _matA0(j, 0) = nearst.points[j].x;
                _matA0(j, 1) = nearst.points[j].y;
                _matA0(j, 2) = nearst.points[j].z;
            }
            _matX0 = _matA0.colPivHouseholderQr().solve(_matB0);

            float pa = _matX0(0, 0);
            float pb = _matX0(1, 0);
            float pc = _matX0(2, 0);
            float pd = 1;

            float ps = std::sqrt(pa * pa + pb * pb + pc * pc);
            pa /= ps;
            pb /= ps;
            pc /= ps;
            pd /= ps;

            bool planeValid = true;
            for (int j = 0; j < 5; j++)
            {
                if (std::fabs(pa * nearst.points[j].x +
                                      pb * nearst.points[j].y +
                                      pc * nearst.points[j].z + pd) > 0.2)
                {
                    planeValid = false;
                    break;
                }
            }

            if (planeValid)
            {
                debug_num22++;
                double dist = pa * _pointSel.x +
                        pb * _pointSel.y +
                        pc * _pointSel.z + pd;
                Eigen::Vector3d omega(pa, pb, pc);
                Eigen::Vector3d point_proj = Eigen::Vector3d(_pointSel.x, _pointSel.y, _pointSel.z) - (dist * omega);
                Eigen::Vector3d e1(1, 0, 0);
                Eigen::Matrix3d J = e1 * omega.transpose();
                Eigen::JacobiSVD<Eigen::Matrix3d> svd(J, Eigen::ComputeThinU | Eigen::ComputeThinV);
                Eigen::Matrix3d R_svd = svd.matrixV() * svd.matrixU().transpose();
                Eigen::Matrix3d info = (1.0 / IMUIntegrator::lidar_m) * Eigen::Matrix3d::Identity();
                info(1, 1) *= plan_weight_tan;
                info(2, 2) *= plan_weight_tan;
                Eigen::Matrix3d sqrt_info = info * R_svd.transpose();

                auto *e = Cost_NavState_IMU_Plan_Vec::Create(Eigen::Vector3d(_pointOri.x, _pointOri.y, _pointOri.z),
                                                             point_proj,
                                                             Tbl,
                                                             sqrt_info);
                edges.push_back(e);
                vPlanFeatures.emplace_back(Eigen::Vector3d(_pointOri.x, _pointOri.y, _pointOri.z),
                                           point_proj,
                                           sqrt_info);
                vPlanFeatures.back().ComputeError(m4d);
            }
        }
    }*/

  }


}


void Estimator::processNonFeatureICP(std::vector<ceres::CostFunction *>& edges,
                                     std::vector<FeatureNon>& vNonFeatures,
                                     const pcl::PointCloud<PointType>::Ptr& laserCloudNonFeature,
                                     const Eigen::Matrix4d& exTlb,
                                     const Eigen::Matrix4d& m4d)
{
  Eigen::Matrix4d Tbl = Eigen::Matrix4d::Identity();
  Tbl.topLeftCorner(3, 3) = exTlb.topLeftCorner(3, 3).transpose();
  Tbl.topRightCorner(3, 1) = -1.0 * Tbl.topLeftCorner(3, 3) * exTlb.topRightCorner(3, 1);
  if (!vNonFeatures.empty())
  {
    for (const auto &p : vNonFeatures)
    {
      auto *e = Cost_NonFeature_ICP::Create(p.pointOri,
                                            p.pa,
                                            p.pb,
                                            p.pc,
                                            p.pd,
                                            Tbl,
                                            Eigen::Matrix<double, 1, 1>(1 / IMUIntegrator::lidar_m));
      edges.push_back(e);
    }
    return;
  }

  PointType _pointOri, _pointSel, _coeff;
  std::vector<int> _pointSearchInd;
  std::vector<float> _pointSearchSqDis;
  std::vector<int> _pointSearchInd2;
  std::vector<float> _pointSearchSqDis2;

  Eigen::Matrix<double, 5, 3> _matA0;
  _matA0.setZero();
  Eigen::Matrix<double, 5, 1> _matB0;
  _matB0.setOnes();
  _matB0 *= -1;
  Eigen::Matrix<double, 3, 1> _matX0;
  _matX0.setZero();

  int laserCloudNonFeatureStackNum = laserCloudNonFeature->points.size();
  for (int i = 0; i < laserCloudNonFeatureStackNum; i++)
  {
    _pointOri = laserCloudNonFeature->points[i];
    LocalMapIvox::pointAssociateToMap(&_pointOri, &_pointSel, m4d);
    if (std::isnan(_pointSel.x) || std::isnan(_pointSel.y) || std::isnan(_pointSel.z)) continue;

    if (m_global_map.laserCloudNonFeatureFromLocal->points.size() > 100)
    {
      m_global_map.kdtreeNonFeatureFromLocal->nearestKSearch(_pointSel, 5, _pointSearchInd, _pointSearchSqDis);
      if (_pointSearchSqDis[4] < 1 * thres_dist)
      {
        for (int j = 0; j < 5; j++)
        {
          _matA0(j, 0) = m_global_map.laserCloudNonFeatureFromLocal->points[_pointSearchInd[j]].x;
          _matA0(j, 1) = m_global_map.laserCloudNonFeatureFromLocal->points[_pointSearchInd[j]].y;
          _matA0(j, 2) = m_global_map.laserCloudNonFeatureFromLocal->points[_pointSearchInd[j]].z;
        }
        _matX0 = _matA0.colPivHouseholderQr().solve(_matB0);

        float pa = _matX0(0, 0);
        float pb = _matX0(1, 0);
        float pc = _matX0(2, 0);
        float pd = 1;

        float ps = std::sqrt(pa * pa + pb * pb + pc * pc);
        pa /= ps;
        pb /= ps;
        pc /= ps;
        pd /= ps;

        bool planeValid = true;
        for (int j = 0; j < 5; j++)
        {
          if (std::fabs(pa * m_global_map.laserCloudNonFeatureFromLocal->points[_pointSearchInd[j]].x +
                                pb * m_global_map.laserCloudNonFeatureFromLocal->points[_pointSearchInd[j]].y +
                                pc * m_global_map.laserCloudNonFeatureFromLocal->points[_pointSearchInd[j]].z + pd) > 0.2)
          {
            planeValid = false;
            break;
          }
        }

        if (planeValid)
        {

          auto *e = Cost_NonFeature_ICP::Create(Eigen::Vector3d(_pointOri.x, _pointOri.y, _pointOri.z),
                                                pa,
                                                pb,
                                                pc,
                                                pd,
                                                Tbl,
                                                Eigen::Matrix<double, 1, 1>(1 / IMUIntegrator::lidar_m));
          edges.push_back(e);
          vNonFeatures.emplace_back(Eigen::Vector3d(_pointOri.x, _pointOri.y, _pointOri.z),
                                    pa,
                                    pb,
                                    pc,
                                    pd);
          vNonFeatures.back().ComputeError(m4d);

          continue;
        }
      }

    }

    /*if (local_map_.NonfMapVox()->NumPoints() > 20)
    {
      pcl::PointCloud<PointType> nearst;
      local_map_.NonfMapVox()->GetClosestPoint(_pointSel, nearst.points, 5, thres_dist);
      if (nearst.size() == 5)
      {
        for (int j = 0; j < 5; j++)
        {
          _matA0(j, 0) = nearst.points[j].x;
          _matA0(j, 1) = nearst.points[j].y;
          _matA0(j, 2) = nearst.points[j].z;
        }
        _matX0 = _matA0.colPivHouseholderQr().solve(_matB0);

        float pa = _matX0(0, 0);
        float pb = _matX0(1, 0);
        float pc = _matX0(2, 0);
        float pd = 1;

        float ps = std::sqrt(pa * pa + pb * pb + pc * pc);
        pa /= ps;
        pb /= ps;
        pc /= ps;
        pd /= ps;

        bool planeValid = true;
        for (int j = 0; j < 5; j++)
        {
          if (std::fabs(pa * nearst.points[j].x +
                                pb * nearst.points[j].y +
                                pc * nearst.points[j].z + pd) > 0.2)
          {
            planeValid = false;
            break;
          }
        }

        if (planeValid)
        {

          auto *e = Cost_NonFeature_ICP::Create(Eigen::Vector3d(_pointOri.x, _pointOri.y, _pointOri.z),
                                                pa,
                                                pb,
                                                pc,
                                                pd,
                                                Tbl,
                                                Eigen::Matrix<double, 1, 1>(1 / IMUIntegrator::lidar_m));
          edges.push_back(e);
          vNonFeatures.emplace_back(Eigen::Vector3d(_pointOri.x, _pointOri.y, _pointOri.z),
                                    pa,
                                    pb,
                                    pc,
                                    pd);
          vNonFeatures.back().ComputeError(m4d);
        }
      }
    }*/
  }
}


void Estimator::vector2double(const std::list<LidarFrame>& lidarFrameList){
  int i = 0;
  for(const auto& l : lidarFrameList)
  {
    Eigen::Map<Eigen::Matrix<double, 6, 1>> PR(para_PR[i]);
    PR.segment<3>(0) = l.P; //前3个值表示坐标
    PR.segment<3>(3) = Sophus::SO3d(l.Q).log();     //后三个值存储旋转向量，表示朝向

    Eigen::Map<Eigen::Matrix<double, 9, 1>> VBias(para_VBias[i]);
    VBias.segment<3>(0) = l.V;              //前3个值表示速度
    VBias.segment<3>(3) = l.bg;             //3-6表示bg
    VBias.segment<3>(6) = l.ba;             //6-9表示ba
    i++;
  }
}

void Estimator::double2vector(std::list<LidarFrame>& lidarFrameList){
  int i = 0;
  for(auto& l : lidarFrameList){
    Eigen::Map<const Eigen::Matrix<double, 6, 1>> PR(para_PR[i]);
    Eigen::Map<const Eigen::Matrix<double, 9, 1>> VBias(para_VBias[i]);
    l.P = PR.segment<3>(0);
    l.Q = Sophus::SO3d::exp(PR.segment<3>(3)).unit_quaternion();
    l.V = VBias.segment<3>(0);
    l.bg = VBias.segment<3>(3);
    l.ba = VBias.segment<3>(6);
    i++;
  }
}

void Estimator::EstimateLidarPose(std::list<LidarFrame>& lidarFrameList,
                           const Eigen::Matrix4d& exTlb,
                           const Eigen::Vector3d& gravity
                           /*nav_msgs::Odometry& debugInfo*/)
{

  //传入当前数据，及前1帧数据，当前帧状态为imu积分得到
  Eigen::Matrix3d exRbl = exTlb.topLeftCorner(3, 3).transpose();  //雷达-imu机械参数变换
  Eigen::Vector3d exPbl = -1.0 * exRbl * exTlb.topRightCorner(3, 1);
  Eigen::Matrix4d transformTobeMapped = Eigen::Matrix4d::Identity();

  //得到imu在世界坐标系下的坐标
  transformTobeMapped.topLeftCorner(3, 3) = lidarFrameList.back().Q * exRbl;
  transformTobeMapped.topRightCorner(3, 1) = lidarFrameList.back().Q * exPbl + lidarFrameList.back().P;

  int laserCloudCornerFromMapNum = m_global_map.laserCloudCornerFromLocal->points.size();
  int laserCloudSurfFromMapNum = m_global_map.laserCloudSurfFromLocal->points.size();


  /*local_map_.CornerMapVox()->GetPoints(local_corner_cloud_.points);
  local_map_.surfMapVox()->GetPoints(local_surf_cloud_.points);
  local_map_.NonfMapVox()->GetPoints(local_nonf_cloud_.points);


  int laserCloudCornerFromLocalNum = local_corner_cloud_.points.size();
  int laserCloudSurfFromLocalNum = local_surf_cloud_.points.size();*/
  int stack_count = 0;

  //遍历2帧历史数据，将角点，平面点，无特征点，提取出，下采样
  for (const auto &l : lidarFrameList)
  {
    laserCloudCornerLast[stack_count]->clear();
    for (const auto &p : l.laserCloud->points)
    {
      if (std::fabs(p.normal_z - 1.0) < 1e-5)
        laserCloudCornerLast[stack_count]->push_back(p);
    }
    laserCloudSurfLast[stack_count]->clear();
    for (const auto &p : l.laserCloud->points)
    {
      if (std::fabs(p.normal_z - 2.0) < 1e-5)
        laserCloudSurfLast[stack_count]->push_back(p);
    }

    laserCloudNonFeatureLast[stack_count]->clear();
    for (const auto &p : l.laserCloud->points)
    {
      if (std::fabs(p.normal_z - 3.0) < 1e-5)
        laserCloudNonFeatureLast[stack_count]->push_back(p);
    }

    laserCloudCornerStack[stack_count]->clear();
    downSizeFilterCorner.setInputCloud(laserCloudCornerLast[stack_count]);
    downSizeFilterCorner.filter(*laserCloudCornerStack[stack_count]);

    laserCloudSurfStack[stack_count]->clear();
    downSizeFilterSurf.setInputCloud(laserCloudSurfLast[stack_count]);
    downSizeFilterSurf.filter(*laserCloudSurfStack[stack_count]);

    laserCloudNonFeatureStack[stack_count]->clear();
    downSizeFilterNonFeature.setInputCloud(laserCloudNonFeatureLast[stack_count]);
    downSizeFilterNonFeature.filter(*laserCloudNonFeatureStack[stack_count]);
    stack_count++;
  }
  //地图有点，执行优化， 更新当前点及历史点的状态
  if (((laserCloudCornerFromMapNum > 0 && laserCloudSurfFromMapNum > 100)/* ||
       (laserCloudCornerFromLocalNum > 0 && laserCloudSurfFromLocalNum > 100)*/))
  {
    Estimate(lidarFrameList, exTlb, gravity);
  }
//根据优化结果重新赋值队列第一帧imu状态
  transformTobeMapped = Eigen::Matrix4d::Identity();
  transformTobeMapped.topLeftCorner(3, 3) = lidarFrameList.front().Q * exRbl;
  transformTobeMapped.topRightCorner(3, 1) = lidarFrameList.front().Q * exPbl + lidarFrameList.front().P;


  //此处加锁 赋值后，地图后台更新线程开始更新地图点
  /*std::unique_lock<std::mutex> locker(mtx_Map);
  *laserCloudCornerForMap = *laserCloudCornerStack[0];              //下标为0，
  *laserCloudSurfForMap = *laserCloudSurfStack[0];
  *laserCloudNonFeatureForMap = *laserCloudNonFeatureStack[0];
  transformForMap = transformTobeMapped;


  //transformTobeMapped 为第一帧数据的imu在世界坐标系下的坐标,更新localmap点云，并下采样（保存，最近50帧）
  //更新局部地图，全局地图更新在后台线程中

  MapIncrementLocal(laserCloudCornerForMap,
                    laserCloudSurfForMap,
                    laserCloudNonFeatureForMap,
                    transformTobeMapped);
  locker.unlock();*/

}

void Estimator::Estimate(std::list<LidarFrame>& lidarFrameList,
                         const Eigen::Matrix4d& exTlb,
                         const Eigen::Vector3d& gravity)
{

  int num_corner_map = 0;
  int num_surf_map = 0;

  static uint32_t frame_count = 0;
  int windowSize = lidarFrameList.size();
  Eigen::Matrix4d transformTobeMapped = Eigen::Matrix4d::Identity();
  Eigen::Matrix3d exRbl = exTlb.topLeftCorner(3, 3).transpose();
  Eigen::Vector3d exPbl = -1.0 * exRbl * exTlb.topRightCorner(3, 1);

  // store point to line features
  std::vector<std::vector<FeatureLine>> vLineFeatures(windowSize);
  for (auto &v : vLineFeatures)
  {
    v.reserve(2000);
  }

  // store point to plan features
  std::vector<std::vector<FeaturePlanVec>> vPlanFeatures(windowSize);
  for (auto &v : vPlanFeatures)
  {
    v.reserve(2000);
  }

  std::vector<std::vector<FeatureNon>> vNonFeatures(windowSize);
  for (auto &v : vNonFeatures)
  {
    v.reserve(2000);
  }

  if (windowSize == SLIDEWINDOWSIZE)
  {
    plan_weight_tan = 0.0003;
    thres_dist = 1.0;
  }
  else
  {
    plan_weight_tan = 0.0;
    thres_dist = 25.0;
  }

  // excute optimize process
  const int max_iters = 10;
  for (int iterOpt = 0; iterOpt < max_iters; ++iterOpt)
  {
    //用freamelist中每帧的P和V赋值优化参数初值，
    vector2double(lidarFrameList);

    //create huber loss function
    ceres::LossFunction *loss_function = NULL;
    loss_function = new ceres::HuberLoss(0.1 / IMUIntegrator::lidar_m);
    if (windowSize == SLIDEWINDOWSIZE)
    {
      loss_function = NULL;
    }
    else
    {
      loss_function = new ceres::HuberLoss(0.1 / IMUIntegrator::lidar_m);
    }

    ceres::Problem::Options problem_options;
    ceres::Problem problem(problem_options);

    //添加每帧数据的位姿优化变量
    for (int i = 0; i < windowSize; ++i)
    {
      //三维P和朝向（rpy）
      problem.AddParameterBlock(para_PR[i], 6);
    }
    //添加每帧数据的速度优化变量
    for (int i = 0; i < windowSize; ++i)
    {
      //三维的V，三维的bg，三维的ba
      problem.AddParameterBlock(para_VBias[i], 9);
    }

    // add IMU CostFunction
    for (int f = 1; f < windowSize; ++f)
    {
      auto frame_curr = lidarFrameList.begin();
      std::advance(frame_curr, f);
      problem.AddResidualBlock(Cost_NavState_PRV_Bias::Create(frame_curr->imuIntegrator,
                                                              const_cast<Eigen::Vector3d &>(gravity),
                                                              Eigen::LLT<Eigen::Matrix<double, 15, 15>>
                                                                      (frame_curr->imuIntegrator.GetCovariance().inverse())
                                                                      .matrixL().transpose()),
                               nullptr,
                               para_PR[f - 1],
                               para_VBias[f - 1],
                               para_PR[f],
                               para_VBias[f]);
    }

    if (last_marginalization_info)
    {
      // construct new marginlization_factor
      auto *marginalization_factor = new MarginalizationFactor(last_marginalization_info);
      problem.AddResidualBlock(marginalization_factor, nullptr,
                               last_marginalization_parameter_blocks);
    }

    Eigen::Quaterniond q_before_opti = lidarFrameList.back().Q;
    Eigen::Vector3d t_before_opti = lidarFrameList.back().P;

    std::vector<std::vector<ceres::CostFunction *>> edgesLine(windowSize);
    std::vector<std::vector<ceres::CostFunction *>> edgesPlan(windowSize);
    std::vector<std::vector<ceres::CostFunction *>> edgesNon(windowSize);
    std::thread threads[2];
    //添加2帧点云与local地图匹配约束

    for (int f = 0; f < windowSize; ++f)
    {

      auto frame_curr = lidarFrameList.begin();
      std::advance(frame_curr, f);
      transformTobeMapped = Eigen::Matrix4d::Identity();
      transformTobeMapped.topLeftCorner(3, 3) = frame_curr->Q * exRbl;
      transformTobeMapped.topRightCorner(3, 1) = frame_curr->Q * exPbl + frame_curr->P;


      TimerRecord::Execute(
              [&,this]() {
                threads[0] = std::thread(&Estimator::processPointToLine, this,
                                         std::ref(edgesLine[f]),
                                         std::ref(vLineFeatures[f]),
                                         std::ref(laserCloudCornerStack[f]),
                                         std::ref(exTlb),
                                         std::ref(transformTobeMapped));


                threads[1] = std::thread(&Estimator::processPointToPlanVec, this,
                                         std::ref(edgesPlan[f]),
                                         std::ref(vPlanFeatures[f]),
                                         std::ref(laserCloudSurfStack[f]),
                                         std::ref(exTlb),
                                         std::ref(transformTobeMapped));


                /*threads[2] = std::thread(&Estimator::processNonFeatureICP, this,
                                         std::ref(edgesNon[f]),
                                         std::ref(vNonFeatures[f]),
                                         std::ref(laserCloudNonFeatureStack[f]),
                                         std::ref(exTlb),
                                         std::ref(transformTobeMapped));*/

                threads[0].join();
                threads[1].join();
                //threads[2].join();
              },"Estimator::FMatch");

    }

    int cntSurf = 0;
    int cntCorner = 0;
    int cntNon = 0;
    if (windowSize == SLIDEWINDOWSIZE)
    {
      thres_dist = 1.0;
      if (iterOpt == 0)
      {
        for (int f = 0; f < windowSize; ++f)
        {
          int cntFtu = 0;
          for (auto &e : edgesLine[f])
          {
            if (std::fabs(vLineFeatures[f][cntFtu].error) > 1e-5)
            {
              problem.AddResidualBlock(e, loss_function, para_PR[f]);
              vLineFeatures[f][cntFtu].valid = true;
            }
            else
            {
              vLineFeatures[f][cntFtu].valid = false;
            }
            cntFtu++;
            cntCorner++;
          }

          cntFtu = 0;
          for (auto &e : edgesPlan[f])
          {
            if (std::fabs(vPlanFeatures[f][cntFtu].error) > 1e-5)
            {
              problem.AddResidualBlock(e, loss_function, para_PR[f]);
              vPlanFeatures[f][cntFtu].valid = true;
            }
            else
            {
              vPlanFeatures[f][cntFtu].valid = false;
            }
            cntFtu++;
            cntSurf++;
          }

          cntFtu = 0;
          for (auto &e : edgesNon[f])
          {
            if (std::fabs(vNonFeatures[f][cntFtu].error) > 1e-5)
            {
              problem.AddResidualBlock(e, loss_function, para_PR[f]);
              vNonFeatures[f][cntFtu].valid = true;
            }
            else
            {
              vNonFeatures[f][cntFtu].valid = false;
            }
            cntFtu++;
            cntNon++;
          }
        }
      }
      else
      {
        for (int f = 0; f < windowSize; ++f)
        {
          int cntFtu = 0;
          for (auto &e : edgesLine[f])
          {
            if (vLineFeatures[f][cntFtu].valid)
            {
              problem.AddResidualBlock(e, loss_function, para_PR[f]);
            }
            cntFtu++;
            cntCorner++;
          }
          cntFtu = 0;
          for (auto &e : edgesPlan[f])
          {
            if (vPlanFeatures[f][cntFtu].valid)
            {
              problem.AddResidualBlock(e, loss_function, para_PR[f]);
            }
            cntFtu++;
            cntSurf++;
          }

          cntFtu = 0;
          for (auto &e : edgesNon[f])
          {
            if (vNonFeatures[f][cntFtu].valid)
            {
              problem.AddResidualBlock(e, loss_function, para_PR[f]);
            }
            cntFtu++;
            cntNon++;
          }
        }
      }
    }
    else
    {
      if (iterOpt == 0)
      {
        thres_dist = 10.0;
      }
      else
      {
        thres_dist = 1.0;
      }
      for (int f = 0; f < windowSize; ++f)
      {
        int cntFtu = 0;
        for (auto &e : edgesLine[f])
        {
          if (std::fabs(vLineFeatures[f][cntFtu].error) > 1e-5)
          {
            problem.AddResidualBlock(e, loss_function, para_PR[f]);
            vLineFeatures[f][cntFtu].valid = true;
          }
          else
          {
            vLineFeatures[f][cntFtu].valid = false;
          }
          cntFtu++;
          cntCorner++;
        }
        cntFtu = 0;
        for (auto &e : edgesPlan[f])
        {
          if (std::fabs(vPlanFeatures[f][cntFtu].error) > 1e-5)
          {
            problem.AddResidualBlock(e, loss_function, para_PR[f]);
            vPlanFeatures[f][cntFtu].valid = true;
          }
          else
          {
            vPlanFeatures[f][cntFtu].valid = false;
          }
          cntFtu++;
          cntSurf++;
        }

        cntFtu = 0;
        for (auto &e : edgesNon[f])
        {
          if (std::fabs(vNonFeatures[f][cntFtu].error) > 1e-5)
          {
            problem.AddResidualBlock(e, loss_function, para_PR[f]);
            vNonFeatures[f][cntFtu].valid = true;
          }
          else
          {
            vNonFeatures[f][cntFtu].valid = false;
          }
          cntFtu++;
          cntNon++;
        }
      }
    }

    ceres::Solver::Options options;
    options.linear_solver_type = ceres::DENSE_SCHUR;
    options.trust_region_strategy_type = ceres::DOGLEG;
    options.max_num_iterations = 10;
    options.minimizer_progress_to_stdout = false;
    options.num_threads = 6;
    ceres::Solver::Summary summary;
  TimerRecord::Execute(
          [&,this]()
          {
            ceres::Solve(options, &problem, &summary);
          },"Estimator::solve"
  );
  final_cost_ = summary.final_cost/float(cntCorner+cntSurf+cntNon);

    //将优化的结果赋值到lidarFrameList的每一帧数据
    double2vector(lidarFrameList);

    Eigen::Quaterniond q_after_opti = lidarFrameList.back().Q;
    Eigen::Vector3d t_after_opti = lidarFrameList.back().P;
    Eigen::Vector3d V_after_opti = lidarFrameList.back().V;
    double deltaR = (q_before_opti.angularDistance(q_after_opti)) * 180.0 / M_PI;
    double deltaT = (t_before_opti - t_after_opti).norm();

    if (deltaR < 0.05 && deltaT < 0.05 || (iterOpt + 1) == max_iters)
    {
      //printf("Frame: %d\n", frame_count++);
      if (windowSize != SLIDEWINDOWSIZE) break;  //此处表示IMU未初始化，
      // apply marginalization
      auto *marginalization_info = new MarginalizationInfo();
      if (last_marginalization_info)
      {
        std::vector<int> drop_set;
        for (int i = 0; i < static_cast<int>(last_marginalization_parameter_blocks.size()); i++)
        {
          if (last_marginalization_parameter_blocks[i] == para_PR[0] ||
                  last_marginalization_parameter_blocks[i] == para_VBias[0])
            drop_set.push_back(i);
        }

        auto *marginalization_factor = new MarginalizationFactor(last_marginalization_info);
        auto *residual_block_info = new ResidualBlockInfo(marginalization_factor, nullptr,
                                                          last_marginalization_parameter_blocks,
                                                          drop_set);
        marginalization_info->addResidualBlockInfo(residual_block_info);
      }

      auto frame_curr = lidarFrameList.begin();
      std::advance(frame_curr, 1);
      ceres::CostFunction *IMU_Cost = Cost_NavState_PRV_Bias::Create(frame_curr->imuIntegrator,
                                                                     const_cast<Eigen::Vector3d &>(gravity),
                                                                     Eigen::LLT<Eigen::Matrix<double, 15, 15>>
                                                                             (frame_curr->imuIntegrator.GetCovariance().inverse())
                                                                             .matrixL().transpose());
      auto *residual_block_info = new ResidualBlockInfo(IMU_Cost, nullptr,
                                                        std::vector<double *>{para_PR[0], para_VBias[0], para_PR[1], para_VBias[1]},
                                                        std::vector<int>{0, 1});
      marginalization_info->addResidualBlockInfo(residual_block_info);

      int f = 0;
      transformTobeMapped = Eigen::Matrix4d::Identity();
      transformTobeMapped.topLeftCorner(3, 3) = frame_curr->Q * exRbl;
      transformTobeMapped.topRightCorner(3, 1) = frame_curr->Q * exPbl + frame_curr->P;
      edgesLine[f].clear();
      edgesPlan[f].clear();
      edgesNon[f].clear();

      TimerRecord::Execute(
              [&,this]() {

                threads[0] = std::thread(&Estimator::processPointToLine, this,
                                         std::ref(edgesLine[f]),
                                         std::ref(vLineFeatures[f]),
                                         std::ref(laserCloudCornerStack[f]),
                                         std::ref(exTlb),
                                         std::ref(transformTobeMapped));

                threads[1] = std::thread(&Estimator::processPointToPlanVec, this,
                                         std::ref(edgesPlan[f]),
                                         std::ref(vPlanFeatures[f]),
                                         std::ref(laserCloudSurfStack[f]),
                                         std::ref(exTlb),
                                         std::ref(transformTobeMapped));

                /*threads[2] = std::thread(&Estimator::processNonFeatureICP, this,
                                         std::ref(edgesNon[f]),
                                         std::ref(vNonFeatures[f]),
                                         std::ref(laserCloudNonFeatureStack[f]),
                                         std::ref(exTlb),
                                         std::ref(transformTobeMapped));*/

                threads[0].join();
                threads[1].join();
                //threads[2].join();
              },"Estimator::FMatch");
      int cntFtu = 0;
      for (auto &e : edgesLine[f])
      {
        if (vLineFeatures[f][cntFtu].valid)
        {
          auto *residual_block_info = new ResidualBlockInfo(e, nullptr,
                                                            std::vector<double *>{para_PR[0]},
                                                            std::vector<int>{0});
          marginalization_info->addResidualBlockInfo(residual_block_info);
        }
        cntFtu++;
      }
      cntFtu = 0;
      for (auto &e : edgesPlan[f])
      {
        if (vPlanFeatures[f][cntFtu].valid)
        {
          auto *residual_block_info = new ResidualBlockInfo(e, nullptr,
                                                            std::vector<double *>{para_PR[0]},
                                                            std::vector<int>{0});
          marginalization_info->addResidualBlockInfo(residual_block_info);
        }
        cntFtu++;
      }

      cntFtu = 0;
      for (auto &e : edgesNon[f])
      {
        if (vNonFeatures[f][cntFtu].valid)
        {
          auto *residual_block_info = new ResidualBlockInfo(e, nullptr,
                                                            std::vector<double *>{para_PR[0]},
                                                            std::vector<int>{0});
          marginalization_info->addResidualBlockInfo(residual_block_info);
        }
        cntFtu++;
      }

      marginalization_info->preMarginalize();
      marginalization_info->marginalize();

      std::unordered_map<long, double *> addr_shift;
      for (int i = 1; i < SLIDEWINDOWSIZE; i++)
      {
        addr_shift[reinterpret_cast<long>(para_PR[i])] = para_PR[i - 1];
        addr_shift[reinterpret_cast<long>(para_VBias[i])] = para_VBias[i - 1];
      }
      std::vector<double *> parameter_blocks = marginalization_info->getParameterBlocks(addr_shift);

      delete last_marginalization_info;
      last_marginalization_info = marginalization_info;
      last_marginalization_parameter_blocks = parameter_blocks;
      break;
    }

    if (windowSize != SLIDEWINDOWSIZE)
    {
      for (int f = 0; f < windowSize; ++f)
      {
        edgesLine[f].clear();
        edgesPlan[f].clear();
        edgesNon[f].clear();
        vLineFeatures[f].clear();
        vPlanFeatures[f].clear();
        vNonFeatures[f].clear();
      }
    }
  }
}

void Estimator::MapIncrementLocal(const pcl::PointCloud<PointType>::Ptr& laserCloudCornerStack,
                                  const pcl::PointCloud<PointType>::Ptr& laserCloudSurfStack,
                                  const pcl::PointCloud<PointType>::Ptr& laserCloudNonFeatureStack,
                                  const Eigen::Matrix4d& transformTobeMapped)
{
    //本地地图特征， 保存50帧最近的点云数据，
    pcl::PointCloud<PointType>::Ptr down_corner(new pcl::PointCloud<PointType>());
    pcl::PointCloud<PointType>::Ptr down_surf(new pcl::PointCloud<PointType>());
    pcl::PointCloud<PointType>::Ptr down_nonf(new pcl::PointCloud<PointType>());
    downSizeFilterCorner.setInputCloud(laserCloudCornerStack);
    downSizeFilterCorner.filter(*down_corner);

    downSizeFilterSurf.setInputCloud(laserCloudSurfStack);
    downSizeFilterSurf.filter(*down_surf);

    downSizeFilterNonFeature.setInputCloud(laserCloudNonFeatureStack);
    downSizeFilterNonFeature.filter(*down_nonf);

  pcl::PointCloud<PointType>::Ptr down_corner_transformed(new pcl::PointCloud<PointType>());
  pcl::PointCloud<PointType>::Ptr down_surf_transformed(new pcl::PointCloud<PointType>());
  pcl::PointCloud<PointType>::Ptr down_nonf_transformed(new pcl::PointCloud<PointType>());
  PointType pointSel;
  PointType pointSel2;
  for (int i = 0; i < down_corner->size(); i++)
  {
    LocalMapIvox::pointAssociateToMap(&laserCloudCornerStack->points[i], &pointSel, transformTobeMapped);
    down_corner_transformed->push_back(pointSel);
  }
  for (int i = 0; i < down_surf->size(); i++)
  {
    LocalMapIvox::pointAssociateToMap(&laserCloudSurfStack->points[i], &pointSel2, transformTobeMapped);
    down_surf_transformed->push_back(pointSel2);
  }
  for (int i = 0; i < down_nonf->size(); i++)
  {
    LocalMapIvox::pointAssociateToMap(&laserCloudNonFeatureStack->points[i], &pointSel2, transformTobeMapped);
    down_nonf_transformed->push_back(pointSel2);
  }


  //local_map_.MapIncrement(down_corner_transformed, down_surf_transformed, down_nonf_transformed, localMapID);


    //本地地图50个点云的汇总点云，清零，然后再更新

    localMapID++;
}