//
// Created by zx on 22-11-10.
//

#include "LocalMapIvox.h"

void LocalMapIvox::Init(int ivox_nearby_type,int max_ivox_node,float resolution)
{
    if (ivox_nearby_type == 0)
    {
        ivox_options_.nearby_type_ = IVoxType::NearbyType::CENTER;
    }
    else if (ivox_nearby_type == 6)
    {
        ivox_options_.nearby_type_ = IVoxType::NearbyType::NEARBY6;
    }
    else if (ivox_nearby_type == 18)
    {
        ivox_options_.nearby_type_ = IVoxType::NearbyType::NEARBY18;
    }
    else if (ivox_nearby_type == 26)
    {
        ivox_options_.nearby_type_ = IVoxType::NearbyType::NEARBY26;
    }
    else
    {
        LOG(WARNING) << "unknown ivox_nearby_type, use NEARBY18";
        ivox_options_.nearby_type_ = IVoxType::NearbyType::NEARBY18;
    }
    ivox_options_.capacity_=max_ivox_node;
    ivox_options_.resolution_=resolution;

    corner_ivox_ = std::make_shared<IVoxType>(ivox_options_);
    surf_ivox_ = std::make_shared<IVoxType>(ivox_options_);
    nonf_ivox_ = std::make_shared<IVoxType>(ivox_options_);

    downSizeFilterCorner.setLeafSize(0.02, 0.02, 0.02);
    downSizeFilterSurf.setLeafSize(0.03, 0.03, 0.03);
    downSizeFilterNonFeature.setLeafSize(0.05, 0.05, 0.05);
}

LocalMapIvox::~LocalMapIvox()
{
}

void LocalMapIvox::MapIncrement(const pcl::PointCloud<PointType>::Ptr& laserCloudCorner,
                            const pcl::PointCloud<PointType>::Ptr& laserCloudSurf,
                            const pcl::PointCloud<PointType>::Ptr& laserCloudNonFeature,int scan_id)
{
    corner_ivox_->AddPoints(laserCloudCorner->points);
    surf_ivox_->AddPoints(laserCloudSurf->points);
    nonf_ivox_->AddPoints(laserCloudNonFeature->points);
    printf(" corner ivox node:%d   surf ivox node:%d  nonf ivox node:%d\n",
            corner_ivox_->NumValidGrids(),surf_ivox_->NumValidGrids(),nonf_ivox_->NumValidGrids());
}

void LocalMapIvox::pointAssociateToMap(PointType const * const pi,
                                   PointType * const po,
                                   const Eigen::Matrix4d& _transformTobeMapped){
    Eigen::Vector3d pin, pout;
    pin.x() = pi->x;
    pin.y() = pi->y;
    pin.z() = pi->z;
    pout = _transformTobeMapped.topLeftCorner(3,3) * pin + _transformTobeMapped.topRightCorner(3,1);
    po->x = pout.x();
    po->y = pout.y();
    po->z = pout.z();
    po->intensity = pi->intensity;
    po->normal_z = pi->normal_z;
}

void LocalMapIvox::featureAssociateToMap(const pcl::PointCloud<PointType>::Ptr& laserCloudCorner,
                                     const pcl::PointCloud<PointType>::Ptr& laserCloudSurf,
                                     const pcl::PointCloud<PointType>::Ptr& laserCloudNonFeature,
                                     const pcl::PointCloud<PointType>::Ptr& laserCloudCornerToMap,
                                     const pcl::PointCloud<PointType>::Ptr& laserCloudSurfToMap,
                                     const pcl::PointCloud<PointType>::Ptr& laserCloudNonFeatureToMap,
                                     const Eigen::Matrix4d& transformTobeMapped){

    int laserCloudCornerNum = laserCloudCorner->points.size();
    int laserCloudSurfNum = laserCloudSurf->points.size();
    int laserCloudNonFeatureNum = laserCloudNonFeature->points.size();
    PointType pointSel1,pointSel2,pointSel3;
    for (int i = 0; i < laserCloudCornerNum; i++) {
        pointAssociateToMap(&laserCloudCorner->points[i], &pointSel1, transformTobeMapped);
        laserCloudCornerToMap->push_back(pointSel1);
    }
    for (int i = 0; i < laserCloudSurfNum; i++) {
        pointAssociateToMap(&laserCloudSurf->points[i], &pointSel2, transformTobeMapped);
        laserCloudSurfToMap->push_back(pointSel2);
    }
    for (int i = 0; i < laserCloudNonFeatureNum; i++) {
        pointAssociateToMap(&laserCloudNonFeature->points[i], &pointSel3, transformTobeMapped);
        laserCloudNonFeatureToMap->push_back(pointSel3);
    }

}
