//
// Created by zx on 22-11-10.
//

#ifndef SRC_LIO_LIVOX_SRC_LIO_LOCALMAPIVOX_H_
#define SRC_LIO_LIVOX_SRC_LIO_LOCALMAPIVOX_H_

#include <pcl/kdtree/kdtree_flann.h>
#include <pcl/point_cloud.h>
#include <pcl/point_types.h>
#include <pcl/filters/voxel_grid.h>
#include <future>
#include "typedef.h"
#include "ivox3d/ivox3d.h"


class LocalMapIvox
{
#ifdef IVOX_NODE_TYPE_PHC
    using IVoxType = faster_lio::IVox<3, faster_lio::IVoxNodeType::PHC, PointType>;
#else
    using IVoxType = faster_lio::IVox<3, faster_lio::IVoxNodeType::DEFAULT, PointType>;
#endif
 public:
    LocalMapIvox()= default;
    void Init(int ivox_nearby_type,int max_ivox_node,float resolution);
    ~LocalMapIvox();

    void MapIncrement(const pcl::PointCloud<PointType>::Ptr& laserCloudCornerStack,
                      const pcl::PointCloud<PointType>::Ptr& laserCloudSurfStack,
                      const pcl::PointCloud<PointType>::Ptr& laserCloudNonFeatureStack,int scan_id);



    std::shared_ptr<IVoxType> CornerMapVox(){
        return corner_ivox_;
    }

    std::shared_ptr<IVoxType> surfMapVox(){
        return surf_ivox_;
    }

    std::shared_ptr<IVoxType> NonfMapVox(){
        return nonf_ivox_;
    }


    static void pointAssociateToMap(PointType const * const pi,
                                    PointType * const po,
                                    const Eigen::Matrix4d& _transformTobeMapped);

    static void featureAssociateToMap(const pcl::PointCloud<PointType>::Ptr& laserCloudCorner,
                                      const pcl::PointCloud<PointType>::Ptr& laserCloudSurf,
                                      const pcl::PointCloud<PointType>::Ptr& laserCloudNonFeature,
                                      const pcl::PointCloud<PointType>::Ptr& laserCloudCornerToMap,
                                      const pcl::PointCloud<PointType>::Ptr& laserCloudSurfToMap,
                                      const pcl::PointCloud<PointType>::Ptr& laserCloudNonFeatureToMap,
                                      const Eigen::Matrix4d& transformTobeMapped);



 protected:
    IVoxType::Options ivox_options_;
    std::shared_ptr<IVoxType> corner_ivox_ = nullptr;
    std::shared_ptr<IVoxType> surf_ivox_ = nullptr;
    std::shared_ptr<IVoxType> nonf_ivox_ = nullptr;

    pcl::VoxelGrid<PointType> downSizeFilterCorner;
    pcl::VoxelGrid<PointType> downSizeFilterSurf;
    pcl::VoxelGrid<PointType> downSizeFilterNonFeature;


};


#endif //SRC_LIO_LIVOX_SRC_LIO_LOCALMAPIVOX_H_
