//
// Created by zx on 22-11-1.
//

#include "mapper.h"
#include "utils/TimerRecord.h"


Mapper::Mapper(int wndsize,LidarFeatureExtractor* feature_extractor,int Use_seg,float filter_coner,float filter_surf,
               int ivox_nearby_type,int max_ivox_node,float resolution)
{
  odomcallback_ = nullptr;
  cloudMappedcallback_ = nullptr;
  final_cost_callback_ = nullptr;
  WINDOWSIZE = wndsize;
  exit_ = false;

  lidarFeatureExtractor_ = feature_extractor;
  Use_seg_ = Use_seg;
  estimator = new Estimator(filter_parameter_corner, filter_parameter_surf,
                            ivox_nearby_type, max_ivox_node, resolution);
  lidarFrameList.reset(new std::list<Estimator::LidarFrame>);

  thread_ = std::thread(&Mapper::process, this);

}

Mapper::~Mapper()
{
    exit_=true;
    if (thread_.joinable())
      thread_.join();
    if(estimator){
        delete estimator;
        estimator= nullptr;
    }
}


void Mapper::completed()
{
    exit_=true;
    if (thread_.joinable())
      thread_.join();
}

void Mapper::InitRT(const Eigen::Matrix3d& R,const Eigen::Vector3d& t)
{
    exTlb.topLeftCorner(3,3) = R;
    exTlb.topRightCorner(3,1) = t;
    exRlb = R;
    exRbl = R.transpose();
    exPlb = t;
    exPbl = -1.0 * exRbl * exPlb;
}

void Mapper::SetFinalCostCallback(CostCallback callback)
{
  final_cost_callback_=callback;
}

void Mapper::SetOdomCallback(odomCallback callback)
{
    odomcallback_=callback;
}

void Mapper::SetCloudMappedCallback(CloudMappedCallback callback)
{
    cloudMappedcallback_=callback;
}

void Mapper::SetLocalMapCloudCallback(CloudMappedCallback callback)
{
    cloudLocalMappcallback_=callback;
}

void Mapper::AddImu(const ImuData& imu)
{
    std::unique_lock<std::mutex> lock(_mutexIMUQueue);
    _imuMsgQueue.push(imu);
}
void Mapper::AddCloud(const TimeCloud& cloud)
{
    std::unique_lock<std::mutex> lock(_mutexLidarQueue);
    _lidarMsgQueue.push(cloud);
}


bool Mapper::fetchImuMsgs(double startTime, double endTime, std::vector<ImuData> &vimuMsg){
    std::unique_lock<std::mutex> lock(_mutexIMUQueue);
    double current_time = 0;
    vimuMsg.clear();
    while(true)
    {
        if(_imuMsgQueue.empty())
            break;
        if(_imuMsgQueue.back().timebase<endTime ||
                _imuMsgQueue.front().timebase>=endTime)
            break;
        ImuData& tmpimumsg = _imuMsgQueue.front();
        double time = tmpimumsg.timebase;
        if(time<=endTime && time>startTime){
            vimuMsg.push_back(tmpimumsg);
            current_time = time;
            _imuMsgQueue.pop();
            if(time == endTime) break;
        } else{
            if(time<=startTime){
                _imuMsgQueue.pop();
            } else{
                double dt_1 = endTime - current_time;
                double dt_2 = time - endTime;
                if(dt_1 < 0 || dt_2 < 0 || dt_1 + dt_2<=0)
                    printf("ERROR  fetchImuMsgs  \n");
                double w1 = dt_2 / (dt_1 + dt_2);
                double w2 = dt_1 / (dt_1 + dt_2);
                //sensor_msgs::ImuPtr theLastIMU(new sensor_msgs::Imu);
                ImuData theLastIMU;
                theLastIMU.linear_acceleration.x = w1 * vimuMsg.back().linear_acceleration.x + w2 * tmpimumsg.linear_acceleration.x;
                theLastIMU.linear_acceleration.y = w1 * vimuMsg.back().linear_acceleration.y + w2 * tmpimumsg.linear_acceleration.y;
                theLastIMU.linear_acceleration.z = w1 * vimuMsg.back().linear_acceleration.z + w2 * tmpimumsg.linear_acceleration.z;
                theLastIMU.angular_velocity.x = w1 * vimuMsg.back().angular_velocity.x + w2 * tmpimumsg.angular_velocity.x;
                theLastIMU.angular_velocity.y = w1 * vimuMsg.back().angular_velocity.y + w2 * tmpimumsg.angular_velocity.y;
                theLastIMU.angular_velocity.z = w1 * vimuMsg.back().angular_velocity.z + w2 * tmpimumsg.angular_velocity.z;
                theLastIMU.timebase=endTime;
                vimuMsg.emplace_back(theLastIMU);
                break;
            }
        }
    }
    return !vimuMsg.empty();
}

/** \brief Remove Lidar Distortion
  * \param[in] cloud: lidar cloud need to be undistorted
  * \param[in] dRlc: delta rotation
  * \param[in] dtlc: delta displacement
  */
void Mapper::RemoveLidarDistortion(pcl::PointCloud<PointType>::Ptr& cloud,
                           const Eigen::Matrix3d& dRlc, const Eigen::Vector3d& dtlc){
    int PointsNum = cloud->points.size();
    for (int i = 0; i < PointsNum; i++) {
        Eigen::Vector3d startP;
        float s = cloud->points[i].normal_x;
        Eigen::Quaterniond qlc = Eigen::Quaterniond(dRlc).normalized();
        Eigen::Quaterniond delta_qlc = Eigen::Quaterniond::Identity().slerp(s, qlc).normalized();
        const Eigen::Vector3d delta_Plc = s * dtlc;
        startP = delta_qlc * Eigen::Vector3d(cloud->points[i].x,cloud->points[i].y,cloud->points[i].z) + delta_Plc;
        Eigen::Vector3d _po = dRlc.transpose() * (startP - dtlc);

        cloud->points[i].x = _po(0);
        cloud->points[i].y = _po(1);
        cloud->points[i].z = _po(2);
        cloud->points[i].normal_x = 1.0;
    }
}

bool Mapper::TryMAPInitialization()
{
//雷达数据达到三帧
    Eigen::Vector3d average_acc = -lidarFrameList->begin()->imuIntegrator.GetAverageAcc();
    double info_g = std::fabs(9.805 - average_acc.norm());
    average_acc = average_acc * 9.805 / average_acc.norm();

    // calculate the initial gravity direction
    double para_quat[4];
    para_quat[0] = 1;
    para_quat[1] = 0;
    para_quat[2] = 0;
    para_quat[3] = 0;

    //优化imu初始坐标系与世界坐标系的旋转变换
    ceres::LocalParameterization *quatParam = new ceres::QuaternionParameterization();
    ceres::Problem problem_quat;

    problem_quat.AddParameterBlock(para_quat, 4, quatParam);

    //初始状态雷达未动， 初始加速度应该=g， imu坐标系与雷达坐标系之间的变换*初始加速度=g，建立误差模型
    problem_quat.AddResidualBlock(Cost_Initial_G::Create(average_acc),
                                  nullptr,
                                  para_quat);

    ceres::Solver::Options options_quat;
    ceres::Solver::Summary summary_quat;
    ceres::Solve(options_quat, &problem_quat, &summary_quat);

    Eigen::Quaterniond q_wg(para_quat[0], para_quat[1], para_quat[2], para_quat[3]);


    //build prior factor of LIO initialization
    Eigen::Vector3d prior_r = Eigen::Vector3d::Zero();
    Eigen::Vector3d prior_ba = Eigen::Vector3d::Zero();
    Eigen::Vector3d prior_bg = Eigen::Vector3d::Zero();
    std::vector<Eigen::Vector3d> prior_v;
    int v_size = lidarFrameList->size();        //前三帧雷达数据
    for (int i = 0; i < v_size; i++)
    {
        prior_v.push_back(Eigen::Vector3d::Zero());
    }
    //对初始旋转四元数做对数变换，得到旋转向量
    Sophus::SO3d SO3_R_wg(q_wg.toRotationMatrix());
    prior_r = SO3_R_wg.log();

    for (int i = 1; i < v_size; i++)
    {
        auto iter = lidarFrameList->begin();
        auto iter_next = lidarFrameList->begin();
        std::advance(iter, i - 1);
        std::advance(iter_next, i);

        Eigen::Vector3d velo_imu = (iter_next->P - iter->P + iter_next->Q * exPlb - iter->Q * exPlb) / (iter_next->timeStamp - iter->timeStamp);
        prior_v[i] = velo_imu;
    }
    prior_v[0] = prior_v[1];

    //3帧数据的3轴速度
    double para_v[v_size][3];
    double para_r[3];
    double para_ba[3];
    double para_bg[3];

    for (int i = 0; i < 3; i++)
    {
        para_r[i] = 0;
        para_ba[i] = 0;
        para_bg[i] = 0;
    }

    for (int i = 0; i < v_size; i++)
    {
        for (int j = 0; j < 3; j++)
        {
            para_v[i][j] = prior_v[i][j];
        }
    }

    Eigen::Matrix<double, 3, 3> sqrt_information_r = 2000.0 * Eigen::Matrix<double, 3, 3>::Identity();
    Eigen::Matrix<double, 3, 3> sqrt_information_ba = 1000.0 * Eigen::Matrix<double, 3, 3>::Identity();
    Eigen::Matrix<double, 3, 3> sqrt_information_bg = 4000.0 * Eigen::Matrix<double, 3, 3>::Identity();
    Eigen::Matrix<double, 3, 3> sqrt_information_v = 4000.0 * Eigen::Matrix<double, 3, 3>::Identity();

    ceres::Problem::Options problem_options;
    ceres::Problem problem(problem_options);
    problem.AddParameterBlock(para_r, 3);
    problem.AddParameterBlock(para_ba, 3);
    problem.AddParameterBlock(para_bg, 3);
    for (int i = 0; i < v_size; i++)
    {
        problem.AddParameterBlock(para_v[i], 3);
    }

    // add CostFunction
    problem.AddResidualBlock(Cost_Initialization_Prior_R::Create(prior_r, sqrt_information_r),
                             nullptr,
                             para_r);

    problem.AddResidualBlock(Cost_Initialization_Prior_bv::Create(prior_ba, sqrt_information_ba),
                             nullptr,
                             para_ba);
    problem.AddResidualBlock(Cost_Initialization_Prior_bv::Create(prior_bg, sqrt_information_bg),
                             nullptr,
                             para_bg);

    for (int i = 0; i < v_size; i++)
    {
        problem.AddResidualBlock(Cost_Initialization_Prior_bv::Create(prior_v[i], sqrt_information_v),
                                 nullptr,
                                 para_v[i]);
    }

    for (int i = 1; i < v_size; i++)
    {
        auto iter = lidarFrameList->begin();
        auto iter_next = lidarFrameList->begin();
        std::advance(iter, i - 1);
        std::advance(iter_next, i);

        Eigen::Vector3d pi = iter->P + iter->Q * exPlb;
        Sophus::SO3d SO3_Ri(iter->Q * exRlb);
        Eigen::Vector3d ri = SO3_Ri.log();
        Eigen::Vector3d pj = iter_next->P + iter_next->Q * exPlb;
        Sophus::SO3d SO3_Rj(iter_next->Q * exRlb);
        Eigen::Vector3d rj = SO3_Rj.log();

        problem.AddResidualBlock(Cost_Initialization_IMU::Create(iter_next->imuIntegrator,
                                                                 ri,
                                                                 rj,
                                                                 pj - pi,
                                                                 Eigen::LLT<Eigen::Matrix<double, 9, 9>>
                                                                         (iter_next->imuIntegrator.GetCovariance().block<
                                                                                 9,
                                                                                 9>(0, 0).inverse())
                                                                         .matrixL().transpose()),
                                 nullptr,
                                 para_r,
                                 para_v[i - 1],
                                 para_v[i],
                                 para_ba,
                                 para_bg);
    }

    ceres::Solver::Options options;
    options.minimizer_progress_to_stdout = false;
    options.linear_solver_type = ceres::DENSE_QR;
    options.num_threads = 6;
    ceres::Solver::Summary summary;
    ceres::Solve(options, &problem, &summary);

    Eigen::Vector3d r_wg(para_r[0], para_r[1], para_r[2]);
    GravityVector = Sophus::SO3d::exp(r_wg) * Eigen::Vector3d(0, 0, -9.805);

    Eigen::Vector3d ba_vec(para_ba[0], para_ba[1], para_ba[2]);
    Eigen::Vector3d bg_vec(para_bg[0], para_bg[1], para_bg[2]);

    if (ba_vec.norm() > 0.5 || bg_vec.norm() > 0.5)
    {
        printf("Too Large Biases! Initialization Failed!\n");
        return false;
    }

    for (int i = 0; i < v_size; i++)
    {
        auto iter = lidarFrameList->begin();
        std::advance(iter, i);
        iter->ba = ba_vec;
        iter->bg = bg_vec;
        Eigen::Vector3d bv_vec(para_v[i][0], para_v[i][1], para_v[i][2]);
        if ((bv_vec - prior_v[i]).norm() > 2.0)
        {
            printf("Too Large Velocity! Initialization Failed!\n");
            std::cout << "delta v norm: " << (bv_vec - prior_v[i]).norm() << std::endl;
            return false;
        }
        iter->V = bv_vec;
    }

    for (size_t i = 0; i < v_size - 1; i++)
    {
        auto laser_trans_i = lidarFrameList->begin();
        auto laser_trans_j = lidarFrameList->begin();
        std::advance(laser_trans_i, i);
        std::advance(laser_trans_j, i + 1);
        laser_trans_j->imuIntegrator.PreIntegration(laser_trans_i->timeStamp, laser_trans_i->bg, laser_trans_i->ba);
    }


    // //if IMU success initialized，保留2帧最近的数据。
    WINDOWSIZE = Estimator::SLIDEWINDOWSIZE;
    while (lidarFrameList->size() > WINDOWSIZE)
    {
        lidarFrameList->pop_front();
    }
    Eigen::Vector3d Pwl = lidarFrameList->back().P;
    Eigen::Quaterniond Qwl = lidarFrameList->back().Q;
    lidarFrameList->back().P = Pwl + Qwl * exPlb;
    lidarFrameList->back().Q = Qwl * exRlb;

    // std::cout << "\n=============================\n| Initialization Successful |"<<"\n=============================\n" << std::endl;

    return true;
}


pcl::PointCloud<PointType>::Ptr Mapper::GetMapCorner()
{
    return estimator->m_global_map.laserCloudCornerFromLocal;
}
pcl::PointCloud<PointType>::Ptr Mapper::GetMapSurf()
{
    return estimator->m_global_map.laserCloudSurfFromLocal;
}
pcl::PointCloud<PointType>::Ptr Mapper::GetMapNonf()
{
    return estimator->m_global_map.laserCloudNonFeatureFromLocal;
}

pcl::PointCloud<PointType>::Ptr Mapper::GetMapFull()
{
    return estimator->m_global_map.full_cloud();
}

void Mapper::SetInitPose(Eigen::Matrix4d transform)
{
    transformAftMapped=transform;
}

/** \brief Mapping main thread
  */
void Mapper::process()
{
  double time_last_lidar = -1;
  double time_curr_lidar = -1;
  Eigen::Matrix3d delta_Rl = Eigen::Matrix3d::Identity();
  Eigen::Vector3d delta_tl = Eigen::Vector3d::Zero();
  Eigen::Matrix3d delta_Rb = Eigen::Matrix3d::Identity();
  Eigen::Vector3d delta_tb = Eigen::Vector3d::Zero();
  std::vector<ImuData> vimuMsg;

  while (exit_ == false)
  {
    auto start = std::chrono::system_clock::now();
    newfullCloud = false;
    pcl::PointCloud<PointType>::Ptr cloudFull(new pcl::PointCloud<PointType>());
    std::unique_lock<std::mutex> lock_lidar(_mutexLidarQueue);
    if (!_lidarMsgQueue.empty())
    {
      // get new lidar msg
      time_curr_lidar = _lidarMsgQueue.front().timebase;
      cloudFull = _lidarMsgQueue.front().cloud;
      _lidarMsgQueue.pop();
      newfullCloud = true;
    }
    lock_lidar.unlock();

    if (newfullCloud == false)
    {
      std::this_thread::sleep_for(std::chrono::milliseconds(10));
      continue;
    }

    if (IMU_Mode > 0 && time_last_lidar > 0)
    {
      // get IMU msg int the Specified time interval
      //第二帧雷达数据以后，进入这里,雷达已经初始化，第一帧雷达imu预积分器使用默认值
      vimuMsg.clear();
      int countFail = 0;
      //将上一帧雷达数据时间点到当前时间点的imu数据从queue中取出
      while (!fetchImuMsgs(time_last_lidar, time_curr_lidar, vimuMsg))
      {
        countFail++;
        if (countFail > 100)
        {
          break;
        }
        std::this_thread::sleep_for(std::chrono::milliseconds(1));
      }

    }
    // this lidar frame init
    Estimator::LidarFrame lidarFrame;
    lidarFrame.laserCloud = cloudFull;
    lidarFrame.timeStamp = time_curr_lidar;

    boost::shared_ptr<std::list<Estimator::LidarFrame>> lidar_list;
    //计算当前雷达数据的imu里程计
    if (!vimuMsg.empty())
    {
      if (!LidarIMUInited)  //imu未初始化（前20帧数据）
      {
        // if get IMU msg successfully, use gyro integration to update delta_Rl
        lidarFrame.imuIntegrator.PushIMUMsg(vimuMsg);
        lidarFrame.imuIntegrator.GyroIntegration(time_last_lidar);
        //imu坐在世界坐标系下的旋转量
        delta_Rb = lidarFrame.imuIntegrator.GetDeltaQ().toRotationMatrix();
        //
        delta_Rl = exTlb.topLeftCorner(3, 3) * delta_Rb * exTlb.topLeftCorner(3, 3).transpose();

        // predict current lidar pose
        lidarFrame.P = transformAftMapped.topLeftCorner(3, 3) * delta_tb
                + transformAftMapped.topRightCorner(3, 1);
        Eigen::Matrix3d m3d = transformAftMapped.topLeftCorner(3, 3) * delta_Rb;
        lidarFrame.Q = m3d;

        lidar_list.reset(new std::list<Estimator::LidarFrame>);
        lidar_list->push_back(lidarFrame);
      }
      else   //imu已经初始化，此时WINDOWSIZE=2，lidarFrameList->size=2
      {
        // if get IMU msg successfully, use pre-integration to update delta lidar pose
        //此时的imuIntegrator为默认imuIntegrator，使用上一帧雷达时刻的imu ba和bg预积分
        lidarFrame.imuIntegrator.PushIMUMsg(vimuMsg);
        lidarFrame.imuIntegrator.PreIntegration(lidarFrameList->back().timeStamp,
                                                lidarFrameList->back().bg,
                                                lidarFrameList->back().ba);

        //上一时刻的雷达状态，（位姿、速度）
        const Eigen::Vector3d &Pwbpre = lidarFrameList->back().P;
        const Eigen::Quaterniond &Qwbpre = lidarFrameList->back().Q;
        const Eigen::Vector3d &Vwbpre = lidarFrameList->back().V;

        //获取当前帧到上一帧的状态差（dp，dq，dv）
        const Eigen::Quaterniond &dQ = lidarFrame.imuIntegrator.GetDeltaQ();
        const Eigen::Vector3d &dP = lidarFrame.imuIntegrator.GetDeltaP();
        const Eigen::Vector3d &dV = lidarFrame.imuIntegrator.GetDeltaV();
        double dt = lidarFrame.imuIntegrator.GetDeltaTime();

        //更新当前帧的状态，赋值当前帧的ba和bg为上一帧的ba、bg。
        lidarFrame.Q = Qwbpre * dQ;
        lidarFrame.P = Pwbpre + Vwbpre * dt + 0.5 * GravityVector * dt * dt + Qwbpre * (dP);
        lidarFrame.V = Vwbpre + GravityVector * dt + Qwbpre * (dV);
        lidarFrame.bg = lidarFrameList->back().bg;
        lidarFrame.ba = lidarFrameList->back().ba;

        //上一雷达时刻imu位姿
        Eigen::Quaterniond Qwlpre = Qwbpre * Eigen::Quaterniond(exRbl);
        Eigen::Vector3d Pwlpre = Qwbpre * exPbl + Pwbpre;

        //当前雷达时刻的imu位姿
        Eigen::Quaterniond Qwl = lidarFrame.Q * Eigen::Quaterniond(exRbl);
        Eigen::Vector3d Pwl = lidarFrame.Q * exPbl + lidarFrame.P;

        //两时刻imu位姿差
        delta_Rl = Qwlpre.conjugate() * Qwl;
        delta_tl = Qwlpre.conjugate() * (Pwl - Pwlpre);
        delta_Rb = dQ.toRotationMatrix();
        delta_tb = dP;

        lidarFrameList->push_back(lidarFrame);
        lidarFrameList->pop_front();
        lidar_list = lidarFrameList;
      }
    }
    else   //imu队列无数据(第一帧雷达数据先到）
    {
      //imu无数据，但已经初始化过， 表示imu出现中断， break，退出程序
      if (LidarIMUInited)
        break;
      else
      {   //第一帧数据
        // predict current lidar pose
        lidarFrame.P = transformAftMapped.topLeftCorner(3, 3) * delta_tb
                + transformAftMapped.topRightCorner(3, 1);
        Eigen::Matrix3d m3d = transformAftMapped.topLeftCorner(3, 3) * delta_Rb;
        lidarFrame.Q = m3d;

        lidar_list.reset(new std::list<Estimator::LidarFrame>);
        lidar_list->push_back(lidarFrame);
      }
    }

    // remove lidar distortion
    //根据imu里程计（当前帧与上一帧的差），消除点云运动畸变
    //特征提取
    pcl::PointCloud<PointType>::Ptr laserConerCloud(new pcl::PointCloud<PointType>);
    pcl::PointCloud<PointType>::Ptr laserSurfCloud(new pcl::PointCloud<PointType>);
    pcl::PointCloud<PointType>::Ptr laserNonFeatureCloud(new pcl::PointCloud<PointType>);
    TimerRecord::Execute(
            [&,this](){

    if (Use_seg_)
    {
      lidarFeatureExtractor_->FeatureExtract_with_segment(cloudFull,
                                                          laserConerCloud,
                                                          laserSurfCloud,
                                                          laserNonFeatureCloud,
                                                          N_SCANS);
    }
    else
    {
      lidarFeatureExtractor_->FeatureExtract(cloudFull, laserConerCloud, laserSurfCloud, N_SCANS);
    }
            },"Feature Extra"
    );

    RemoveLidarDistortion(cloudFull, delta_Rl, delta_tl);

    TimerRecord::Execute(
            [&,this](){


    estimator->EstimateLidarPose(*lidar_list, exTlb, GravityVector);

            },"ceres"
    );
    pcl::PointCloud<PointType>::Ptr laserCloudCornerMap(new pcl::PointCloud<PointType>());
    pcl::PointCloud<PointType>::Ptr laserCloudSurfMap(new pcl::PointCloud<PointType>());


    Eigen::Matrix4d transformTobeMapped = Eigen::Matrix4d::Identity();
    transformTobeMapped.topLeftCorner(3, 3) = lidar_list->front().Q * exRbl;
    transformTobeMapped.topRightCorner(3, 1) = lidar_list->front().Q * exPbl + lidar_list->front().P;

    // update delta transformation
    delta_Rb = transformAftMapped.topLeftCorner(3, 3).transpose() * lidar_list->front().Q.toRotationMatrix();
    delta_tb = transformAftMapped.topLeftCorner(3,
                                                3).transpose() * (lidar_list->front().P - transformAftMapped.topRightCorner(
            3,
            1));
    Eigen::Matrix3d Rwlpre = transformAftMapped.topLeftCorner(3, 3) * exRbl;
    Eigen::Vector3d Pwlpre = transformAftMapped.topLeftCorner(3, 3) * exPbl + transformAftMapped.topRightCorner(
            3,
            1);
    delta_Rl = Rwlpre.transpose() * transformTobeMapped.topLeftCorner(3, 3);
    delta_tl = Rwlpre.transpose() * (transformTobeMapped.topRightCorner(3, 1) - Pwlpre);
    transformAftMapped.topLeftCorner(3, 3) = lidar_list->front().Q.toRotationMatrix();
    transformAftMapped.topRightCorner(3, 1) = lidar_list->front().P;

    // publish odometry rostopic
    if (odomcallback_ != nullptr)
    {
      odomcallback_(transformTobeMapped, lidar_list->front().timeStamp);
    }

    // publish lidar points
    if (cloudMappedcallback_ != nullptr)
    {
      int laserCloudFullResNum = lidar_list->front().laserCloud->points.size();
      pcl::PointCloud<PointType>::Ptr laserCloudAfterEstimate(new pcl::PointCloud<PointType>());
      laserCloudAfterEstimate->reserve(laserCloudFullResNum);
      for (int i = 0; i < laserCloudFullResNum; i++)
      {
        PointType temp_point;
        LocalMapIvox::pointAssociateToMap(&lidar_list->front().laserCloud->points[i],
                                          &temp_point,
                                          transformTobeMapped);
        laserCloudAfterEstimate->push_back(temp_point);
      }
      cloudMappedcallback_(laserCloudAfterEstimate, lidar_list->front().timeStamp);
    }

    /*if (cloudLocalMappcallback_ != nullptr)
    {

      pcl::PointCloud<PointType> all;

      all += estimator->local_corner_cloud_;
      all += estimator->local_surf_cloud_;
      all += estimator->local_nonf_cloud_;

      cloudLocalMappcallback_(all.makeShared(), lidar_list->front().timeStamp);
    }*/

    if (final_cost_callback_ != nullptr)
    {
      final_cost_callback_(estimator->final_cost_);
    }

    // if tightly coupled IMU message, start IMU initialization
    if (IMU_Mode > 1 && !LidarIMUInited)
    {
      // update lidar frame pose
      lidarFrame.P = transformTobeMapped.topRightCorner(3, 1);
      Eigen::Matrix3d m3d = transformTobeMapped.topLeftCorner(3, 3);
      lidarFrame.Q = m3d;

      // static int pushCount = 0;
      if (pushCount == 0)
      {
        lidarFrameList->push_back(lidarFrame);
        lidarFrameList->back().imuIntegrator.Reset();
        if (lidarFrameList->size() > WINDOWSIZE)
          lidarFrameList->pop_front();
      }
      else
      {
        lidarFrameList->back().laserCloud = lidarFrame.laserCloud;
        lidarFrameList->back().imuIntegrator.PushIMUMsg(vimuMsg);
        lidarFrameList->back().timeStamp = lidarFrame.timeStamp;
        lidarFrameList->back().P = lidarFrame.P;
        lidarFrameList->back().Q = lidarFrame.Q;
      }
      pushCount++;
      if (pushCount >= 3)
      {
        pushCount = 0;
        if (lidarFrameList->size() > 1)
        {
          auto iterRight = std::prev(lidarFrameList->end());
          auto iterLeft = std::prev(std::prev(lidarFrameList->end()));
          iterRight->imuIntegrator.PreIntegration(iterLeft->timeStamp, iterLeft->bg, iterLeft->ba);
        }
        //记录前2/3 windowsize的数据时间点， 此时间点之前的数据不参与imu初始化，初始化完成后置WINDOWSIE=2
        if (lidarFrameList->size() == int(WINDOWSIZE / 1.5))
        {
          startTime = lidarFrameList->back().timeStamp;
        }

        if (!LidarIMUInited && lidarFrameList->size() == WINDOWSIZE && lidarFrameList->front().timeStamp >= startTime)
        {
          std::cout << "**************Start MAP Initialization!!!******************" << std::endl;
          if (TryMAPInitialization())
          {
            LidarIMUInited = true;
            pushCount = 0;
            startTime = 0;
          }
          std::cout << "**************Finish MAP Initialization!!!******************" << std::endl;
        }

      }
    }
    time_last_lidar = time_curr_lidar;
    auto end = std::chrono::system_clock::now();
    auto duration = std::chrono::duration_cast<std::chrono::microseconds>(end - start);
    frame_tm_ = double(duration.count()) * std::chrono::microseconds::period::num / std::chrono::microseconds::period::den;

  }
}

void Mapper::LoadMap(std::string file)
{
    estimator->m_global_map.load_map(file);
}
