//
// Created by zx on 22-11-1.
//

#ifndef SRC_LIO_LIVOX_SRC_LIO_MAPPER_H_
#define SRC_LIO_LIVOX_SRC_LIO_MAPPER_H_

#include "Estimator.h"

#include "typedef.h"
#include "segment/LidarFeatureExtractor.h"
#include <pcl/registration/ndt.h>

class Mapper
{
 public:
    Mapper(int wndsize,LidarFeatureExtractor* feature_extractor,int Use_seg,float filter_coner,float filter_surf,
           int ivox_nearby_type,int max_ivox_node,float resolution);
    ~Mapper();
    void InitRT(const Eigen::Matrix3d& R,const Eigen::Vector3d& T);
    void AddImu(const ImuData& imu);
    void AddCloud(const TimeCloud& cloud);


    void SetOdomCallback(odomCallback callback);
    void SetCloudMappedCallback(CloudMappedCallback callback);
    void SetFinalCostCallback(CostCallback callback);

    void SetLocalMapCloudCallback(CloudMappedCallback callback);

    bool fetchImuMsgs(double startTime, double endTime, std::vector<ImuData> &vimuMsg);

    void RemoveLidarDistortion(pcl::PointCloud<PointType>::Ptr& cloud,
                               const Eigen::Matrix3d& dRlc, const Eigen::Vector3d& dtlc);

    bool TryMAPInitialization();

    void process();
    void completed();
    void LoadMap(std::string file);
    pcl::PointCloud<PointType>::Ptr GetMapFull();
    pcl::PointCloud<PointType>::Ptr GetMapCorner();
    pcl::PointCloud<PointType>::Ptr GetMapSurf();
    pcl::PointCloud<PointType>::Ptr GetMapNonf();
    void SetInitPose(Eigen::Matrix4d transform);


 public:
    int WINDOWSIZE;

    bool exit_;
    std::thread thread_;

    odomCallback odomcallback_;
    CloudMappedCallback cloudMappedcallback_;
    CloudMappedCallback cloudLocalMappcallback_;
    CostCallback final_cost_callback_;


    bool LidarIMUInited = false;
    boost::shared_ptr<std::list<Estimator::LidarFrame>> lidarFrameList;
    LidarFeatureExtractor* lidarFeatureExtractor_;
    Estimator* estimator;

    bool newfullCloud = false;

    Eigen::Matrix4d transformAftMapped = Eigen::Matrix4d::Identity();

    std::mutex _mutexLidarQueue;

    std::queue<TimeCloud> _lidarMsgQueue;
    std::mutex _mutexIMUQueue;
    std::queue<ImuData> _imuMsgQueue;
    Eigen::Matrix4d exTlb;      //雷达到imu变换
    Eigen::Matrix3d exRlb, exRbl;
    Eigen::Vector3d exPlb, exPbl;
    Eigen::Vector3d GravityVector;
    float filter_parameter_corner = 0.2;
    float filter_parameter_surf = 0.4;
    int IMU_Mode = 2;
    int pushCount = 0;
    double startTime = 0;
    int N_SCANS=6;
    int Use_seg_=true;
    double frame_tm_=0;

};


#endif //SRC_LIO_LIVOX_SRC_LIO_MAPPER_H_
