//
// The MIT License (MIT)
//
// Copyright (c) 2019 Livox. All rights reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.
//
#ifndef LIVOX_ROS_DRIVER_LDDC_H_
#define LIVOX_ROS_DRIVER_LDDC_H_

#include "lds.h"
#include "livox_sdk.h"

#include "typedef.h"
#include <thread>
namespace livox_ros {

typedef void (*PointCloudCallback)(TimeCloud,int);
typedef void (*ImuCallback)(ImuData,int);

/** Lidar data distribute control */
typedef enum {
  kPointCloud2Msg = 0,
  kLivoxCustomMsg = 1,
  kPclPxyziMsg
} TransferType;

class Lddc {
 public:
  Lddc(double frq);
  ~Lddc();

  int RegisterLds(Lds *lds);
  void DistributeLidarData(void);

  void SetCloudCallback(PointCloudCallback callback){
      private_pub_=callback;
  }
  void SetImuCallback(ImuCallback callback)
  {
      private_imu_pub_=callback;
  }

  void Start();
  void Start(std::string lvx);
  void Stop();
  void PrepareExit(void);


  void SetPublishFrq(uint32_t frq) {
    if(exit_==false)
      return ;
    publish_frq_ = frq;
    publish_period_ns_ = kNsPerSecond / publish_frq_;
  }

  Lds *lds_;

 private:
  int32_t GetPublishStartTime(LidarDevice *lidar, LidarDataQueue *queue,
                              uint64_t *start_time,
                              StoragePacket *storage_packet);
  uint32_t PublishCustomPointcloud(LidarDataQueue *queue, uint32_t packet_num,
                                   uint8_t handle);
  uint32_t PublishImuData(LidarDataQueue *queue, uint32_t packet_num,
                          uint8_t handle);

  static void capture(Lddc* p);


  void PollingLidarPointCloudData(uint8_t handle, LidarDevice *lidar);
  void PollingLidarImuData(uint8_t handle, LidarDevice *lidar);
  void FillPointsToCustomMsg(TimeCloud& livox_msg, \
      LivoxPointXyzrtl* src_point, uint32_t num, uint32_t offset_time, \
      uint32_t point_interval, uint32_t echo_num);
  double publish_frq_;
  uint32_t publish_period_ns_;

    PointCloudCallback private_pub_;
    ImuCallback private_imu_pub_;

    std::string lvx_file_;
    LidarDataSourceType dataSourceType_=kSourceRawLidar;
    std::thread* capture_;
    bool exit_=false;


};

}  // namespace livox_ros
#endif
